import numba as nb
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path
import svcco
import pyvista as pv
from tqdm import trange, tqdm
import numpy as np
import tetgen
from scipy import stats
import pymeshfix
import pickle
import matplotlib.pyplot as plt
import matplotlib as mpl
from tempfile import TemporaryDirectory
import os
from multiprocessing import Pool

def get_all_paths(i,graph):
    ds, Pr = shortest_path(csgraph=graph,directed=False,indices=i,return_predecessors=True)
    return ds, Pr

def parallel_write(pkg):
    start   = pkg[0]
    stop    = pkg[1]
    tempdir = pkg[2]
    graph   = pkg[3]
    position = pkg[4]
    graph_range = list(range(start,stop))
    file_ds = np.memmap(tempdir+os.sep+'ds_{}-{}.dat'.format(start,stop),dtype=np.float64,mode='w+',shape=(len(graph_range),graph.shape[1]))
    file_Pr = np.memmap(tempdir+os.sep+'Pr_{}-{}.dat'.format(start,stop),dtype=np.int64,mode='w+',shape=(len(graph_range),graph.shape[1]))
    #tqdm(range(len(data['flow'])),desc="Building Vessel Data",position=1,leave=False)
    #graph_range = list(range(start,stop))
    checks = np.linspace(0,len(graph_range),10,dtype=int)
    for i in range(len(graph_range)):
        if i in checks:
            print("Dijkstra Solve | {} - {}:  {}%".format(start,stop,i/len(graph_range)*100))
        ds,Pr = get_all_paths(graph_range[i],graph)
        file_ds[i,:] = ds
        file_Pr[i,:] = Pr
        file_ds.flush()
        file_Pr.flush()
        del ds
        del Pr
    print("Dijkstra Solve | {} - {}:  {}%".format(start,stop,100))
    del file_ds
    del file_Pr
    return

def determine_blocks(row_number,block_number):
    base = row_number//block_number
    remain = row_number%block_number
    blocks = []
    count  = 0
    for i in range(block_number):
        tmp = [count]
        tmp.append(count+base)
        if i == block_number - 1:
            tmp[-1] += remain
        blocks.append(tmp)
        count += base
    return blocks

def build_work(blocks,tmpdir,graph):
    work = []
    for i,b in enumerate(blocks):
        b.append(tmpdir.name)
        b.append(graph)
        b.append(i)
        work.append(b)
    work = tuple(work)
    return work

def precompute_graph_predecessors(surface_object,block_num=10,parallel_num=10):
    file_dir  = 'D:'+os.sep
    tmp_dir   = TemporaryDirectory(dir=file_dir)
    #print(tmp_dir.name)
    surface_object.tmp_dir_dijkstra = tmp_dir
    #surface_object.graph_predecessor_distance = np.memmap(tmp_dir.name + os.sep + 'ds.dat',dtype='int32',mode='w+',shape=surface_object.graph.shape)
    #surface_object.graph_predecessor_id       = np.memmap(tmp_dir.name + os.sep + 'Pr.dat',dtype='int32',mode='w+',shape=surface_object.graph.shape)
    surface_object.cell_centers = surface_object.tet.grid.cell_centers().points
    surface_object.cell_center_closest_point = []
    #for i in trange(surface_object.tet.grid.n_points):
    #    ds,Pr = get_all_paths(i,surface_object.graph)
    #    surface_object.graph_predecessor_distance[i,:] = ds
    #    surface_object.graph_predecessor_id[i,:]       = Pr
    #    del ds
    #    del Pr
    blocks = determine_blocks(surface_object.graph.shape[0],block_num)
    base = blocks[0][1] - blocks[0][0]
    work = build_work(blocks,tmp_dir,surface_object.graph)
    p = Pool(parallel_num)
    p.map(parallel_write,work)
    ds = {'ranges':[],'files':[],'mats':[]}
    Pr = {'ranges':[],'files':[],'mats':[]}
    for i in range(len(work)):
         ds['ranges'].append(work[i][0:2])
         ds['files'].append(tmp_dir.name+os.sep+'ds_{}-{}.dat'.format(work[i][0],work[i][1]))
         ds['mats'].append(np.memmap(tmp_dir.name+os.sep+'ds_{}-{}.dat'.format(work[i][0],work[i][1]),mode='r',dtype=np.float64,shape=(work[i][1]-work[i][0],surface_object.graph.shape[1])))
         Pr['ranges'].append(work[i][0:2])
         Pr['files'].append(tmp_dir.name+os.sep+'Pr_{}-{}.dat'.format(work[i][0],work[i][1]))
         Pr['mats'].append(np.memmap(tmp_dir.name+os.sep+'Pr_{}-{}.dat'.format(work[i][0],work[i][1]),mode='r',dtype=np.int64,shape=(work[i][1]-work[i][0],surface_object.graph.shape[1])))
    surface_object.ds = ds
    surface_object.Pr = Pr
    tet_vert_map = []
    for i in trange(surface_object.tet.grid.n_cells):
        surface_object.cell_center_closest_point.append(surface_object.tet.grid.find_closest_point(surface_object.cell_centers[i,:]))
        if surface_object.cell_center_closest_point[-1] < blocks[-1][0]:
            tmp = [surface_object.cell_center_closest_point[-1]//base,surface_object.cell_center_closest_point[-1]%base]
        else:
            tmp = [len(blocks)-1,surface_object.cell_center_closest_point[-1] - blocks[-1][0]]
        tet_vert_map.append(tmp)
    surface_object.tet_vert_map = tet_vert_map
    return

@nb.jit(nopython=True)
def get_path(j,ds,Pr):
    path = [j]
    lengths = []
    k = j
    while Pr[k] != -9999:
        path.append(Pr[k])
        lengths.append(ds[k])
        k = Pr[k]
    path = path[::-1]
    lengths = lengths[::-1]
    lines = []
    for jdx in range(len(path)-1):
        lines.append([path[jdx],path[jdx+1]])
    return path,lengths,lines

def perfusion_geodesic(tree,perfusion_volume):
    terminals = tree.data[tree.data[:,15]<0,:]
    terminals = terminals[terminals[:,16]<0,:]
    terminal_ids = []
    vol = perfusion_volume.tet.grid.compute_cell_sizes().cell_data['Volume']
    for idx in range(terminals.shape[0]):
        terminal_ids.append(perfusion_volume.tet.grid.find_closest_point(terminals[idx,3:6]))
    territory_id = []
    territory_volumes = np.zeros(terminals.shape[0])
    for idx in tqdm(range(perfusion_volume.tet.grid.n_cells),desc='Calculating Perfusion Territories'):
        tmp_geodesic_lengths = []
        linear_distances     = []
        center = perfusion_volume.cell_centers[idx,:] #perfusion_volume.tet.grid.cell_centers().points[idx,:]
        cell_id = perfusion_volume.cell_center_closest_point[idx] #perfusion_volume.tet.grid.find_closest_point(center)
        for jdx in range(len(terminal_ids)):
            #tmp_ds = perfusion_volume.graph_predecessor_distan[cell_id]
            #tmp_Pr = perfusion_volume.graph_predecessor_id[cell_id]
            tmp_ds = np.array(perfusion_volume.ds['mats'][perfusion_volume.tet_vert_map[idx][0]][perfusion_volume.tet_vert_map[idx][1],:])
            tmp_Pr = np.array(perfusion_volume.Pr['mats'][perfusion_volume.tet_vert_map[idx][0]][perfusion_volume.tet_vert_map[idx][1],:])
            _,L,_ = get_path(terminal_ids[jdx],tmp_ds,tmp_Pr)
            L = sum(L)
            tmp_geodesic_lengths.append(L)
            linear_distances.append(np.linalg.norm(terminals[jdx,3:6] - center))
        tmp_geodesic_lengths = np.array(tmp_geodesic_lengths)
        min_geodesic_value = np.min(tmp_geodesic_lengths)
        min_geodesic_instances = np.where(tmp_geodesic_lengths==min_geodesic_value)
        linear_distances = np.array(linear_distances)
        if len(min_geodesic_instances) == 1:
            territory_id.append(min_geodesic_instances[0])
            territory_volumes[min_geodesic_instances[0]] += vol[idx]
        else:
            absolute_min = np.argwhere(linear_distances[min_geodesic_instances])
            absolute_min = min_geodesic_instances[absolute_min]
            territory_id.append(absolute_min)
            territory_volumes[absolute_min] += vol[idx]
    territory_id = np.array(territory_id)
    perfusion_volume.tet.grid['perfusion_territory_id'] = territory_id
    return territory_id,territory_volumes/np.sum(vol),perfusion_volume.tet

def test(surf_object,size=10,restarts=10):
    t = svcco.tree()
    t.set_boundary(surf_object)
    delaunay = surf_object.pv_polydata.delaunay_3d()
    convexity = surf_object.volume/delaunay.volume
    if convexity > 0.95:
        t.convex = True
    t.set_root()
    add_amount = size//restarts
    VOLS  = []
    COUNT = []
    _,vols,_ = perfusion_geodesic(t,surf_object)
    for i in range(restarts):
        t.n_add(add_amount)
        if i == 0:
            _,vols,tet = perfusion_geodesic(t,surf_object)
        else:
            _,vols,tet = perfusion_geodesic(t,surf_object)
        VOLS.append(vols)
        if i == 0:
            COUNT.append(add_amount)
        else:
            COUNT.append(COUNT[-1]+add_amount)
    return VOLS,COUNT

def results(test_list,size=10,restarts=10,bins=50,repeat=1):
    fig,ax = plt.subplots(nrows=2,ncols=2)
    #ax = ax.flatten()
    #cw = plt.get_cmap('coolwarm')
    norm = mpl.colors.Normalize(vmin=0, vmax=size+1)
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.coolwarm)
    cmap.set_array([])
    DATA = {'volumes':[],'tree_size':[]}
    for i in range(len(test_list)):
        vols,sizes = test(test_list[i],size=size,restarts=restarts)
        DATA['volumes'].append(vols)
        DATA['tree_size'].append(sizes)
        DATA['repeat'] = repeat
        idx = i//2
        jdx = i%2
        for j in range(len(vols)):
            #color_scale = (sizes[j]/size)*255
            x_min = min(vols[-1])
            x_max = max(vols[-1])
            x_range = np.linspace(x_min,x_max)
            clean_vols = vols[j][vols[j]>0]
            freq,edges = np.histogram(clean_vols,range=(0,1),bins=bins)
            centers = 0.5*(edges[1:]+edges[:-1])
            width   = edges[1]-edges[0]
            #if j == 0:
            #    ax[idx][jdx].bar(centers,freq/len(vols[j]),width=width,color='blue')
            #else:
            ax[idx][jdx].bar(centers,freq/len(clean_vols),width=width,alpha=0.4,color=cmap.to_rgba(len(vols[j])))
            #ax[idx][jdx].hist(vols[j],bins=50,density=True,alpha=0.25,color=cw(color_scale),range = (x_min,x_max))
            #kde = stats.gaussian_kde(vols[j])
            #ax[idx][jdx].plot(x_range,kde(x_range)/10,color=cw(color_scale),alpha=0.75)
    #sm = plt.cm.ScalarMappable(cmap='coolwarm')
    #sm.set_array(range(0,size))
    if repeat > 1:
        for j in range(len(test_list)):
            for i in range(len(DATA['volumes'][j])):
                DATA['volumes'][j][i] = DATA['volumes'][j][i].tolist()
        for j in range(repeat-1):
            for i in range(len(test_list)):
                vols,sizes = test(test_list[i],size=size,restarts=restarts)
                for k in range(len(vols)):
                    DATA['volumes'][i][k].extend(vols[k])
    fig.colorbar(cmap,ax=ax.ravel().tolist(),label="Number of Terminals")
    fig.savefig('perfusion-{}_num_vessels-{}_restarts-{}_num_bins-{}.svg'.format(len(test_list),size,restarts,bins),format='svg')
    return fig,ax,DATA

"""
q = 4
resolution = 20 #120

cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve(quiet=False)
cube.build(q=q,resolution=resolution,verbose=True)
print('preprocessing cells')

blocks = determine_blocks(cube.graph.shape[0],10)
tmpdir = TemporaryDirectory(dir='D:'+os.sep)
work = build_work(blocks,tmpdir,cube.graph)



precompute_graph_predecessors(cube)
print('cube constructed')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve(quiet=False)
heart.build(q=4,resolution=120,k=2,buffer=5,verbose=True)
print('preprocessing cells')
precompute_graph_predecessors(heart)
print('heart constructed')


disk = pv.Disc(inner=2.8,outer=4,r_res=20,c_res=100)
cyl  = disk.extrude([0,0,2],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve(quiet=False)
cylinder.build(q=q,resolution=resolution,verbose=True)
print('preprocessing cells')
precompute_graph_predecessors(cylinder)
print('cylinder constructed')

left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus_no_scale = pv.read(left_gyrus)
sf = (heart.volume/gyrus_no_scale.volume)**(1/3)
gyrus_scaled = gyrus_no_scale.scale([sf,sf,sf])
left_gyrus_scaled = "left_gyrus_scaled.vtp"
gyrus_scaled.save(left_gyrus_scaled)
gyrus = svcco.surface()
gyrus.load(left_gyrus_scaled)
gyrus.solve(quiet=False)
gyrus.build(q=q,resolution=resolution,buffer=5,verbose=True)
print('preprocessing cells')
precompute_graph_predecessors(gyrus)
print('gyrus constructed')

f,a,data = results([cube,cylinder,heart,gyrus],size=50,restarts=50,repeat=10)
file = open('perfusion_fast_data.pkl','wb+')
pickle.dump(data,file)
file.close()
"""



# File to test tangent constrained splines

import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate, optimize
from mpl_toolkits.mplot3d import Axes3D
import numba as nb
import nlopt

def test():
    P1 = np.random.random(3)
    P2 = np.random.random(3)
    V1 = P2-P1
    # Tangent Constraint 1 for First Spline
    V1 = V1/np.linalg.norm(V1)
    P3 = np.random.random(3)
    P4 = np.random.random(3)
    V2 = P4-P3
    # Tangent Constraint 1 for Second Spline
    V2 = V2/np.linalg.norm(V2)
    #MID = (P2 + P4)/2
    #V3  = MID - P2
    # Tangent Constraint 2 for First Spline
    V3  = V3/np.linalg.norm(V3)
    #V4  = MID - P4
    # Tangent Constraint 2 for Second Spline
    V4  = V4/np.linalg.norm(V4)
    CTR0 = P2 + V1*0.01
    CTR1 = P2 + V1*0.1
    CTR2 = P4 - V3*0.1
    #CTR3 = P4 - V3*0.01
    #CTR4 = MID + V3*0.25
    #CTR3 = P4 + V2*0.1
    # Create Cubic Spline 1
    SPT1 = np.vstack((P2,CTR0,CTR1,CTR2,CTR3,MID))
    tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=3,nest=-1)
    tck2, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=3)
    l, r = [(1, (0, 0, 0))], [(1, (0, 0, 0))]
    clamped_spline = interpolate.make_interp_spline(u,SPT1,bc_type=(l,r))
    spline_f1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=0))
    spline_df1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=1)).flatten()/np.linalg.norm(np.array(interpolate.splev(t,tck=tck1,der=1)).flatten())
    print('True Vector Inlet:   {}'.format(V1))
    print('Spline Vector Inlet: {}'.format(spline_df1(0)))
    fig = plt.figure()
    ax = fig.add_subplot(111,projection='3d')
    tt = np.linspace(0,1,num=400)
    spline1 = spline_f1(tt).T
    spline2 = clamped_spline(tt)
    ax.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='unclamped')
    ax.plot(spline2[:,0],spline2[:,1],spline2[:,2],color='red',label='clamped')
    ax.plot([P1[0],P2[0]],[P1[1],P2[1]],[P1[2],P2[2]],color='green',label='inlet')
    ax.plot([MID[0],CTR4[0]],[MID[1],CTR4[1]],[MID[2],CTR4[2]],color='green',label='outlet')
    ax.scatter(SPT1[:,0],SPT1[:,1],SPT1[:,2],color='black',label='data')
    ax.legend()
    plt.show()

@nb.jit(nopython=True)
def get_angle(V1,V2):
    return np.arccos(np.dot(-V1,V2))*(180/np.pi)


@nb.jit(nopython=True)
def get_vecs(vec):
    angles = np.zeros(vec.shape[0]-1)
    for i in range(vec.shape[0]-1):
        value = get_angle(vec[i,:],vec[i+1,:])
        angles[i] = 1/value
    return angles


def curvature(t,df=None,ddf=None):
    numerator = np.cross(df(t),ddf(t))
    kappa = np.linalg.norm(numerator)/np.linalg.norm(df(t))**3
    return kappa

collision_vessels = np.array([[0,0.5,0.25,1,0.5,0.25,0.01],
                              [0,0.7,0.25,1,0.7,0.25,0.01],
                              [0,0.3,0.25,1,0.3,0.25,0.01],
                              [0,0.9,0.25,1,0.9,0.25,0.01],
                              [0,0.1,0.25,1,0.1,0.25,0.01],
                              [0.25,0,0.25,0.25,1,0.25,0.01],
                              [0.75,0,0.25,0.75,1,0.25,0.01],
                              [0,0.5,0.75,1,0.5,0.75,0.01],
                              [0,0.7,0.75,1,0.7,0.75,0.01],
                              [0,0.3,0.75,1,0.3,0.75,0.01],
                              [0,0.9,0.75,1,0.9,0.75,0.01],
                              [0,0.1,0.75,1,0.1,0.75,0.01],
                              [0.25,0,0.75,0.25,1,0.75,0.01],
                              [0.75,0,0.75,0.75,1,0.75,0.01],
                              [1,0,0,1,0,1,0.1]])


pt = np.array([0,0,0])
pts= np.random.random((3,3))

from math import tanh

@nb.jit(nopython=True)
def get_collisions(collision_vessels,R,sample_pts):
    collisions = 0
    for i in range(sample_pts.shape[0]):
        dist = close_exact(collision_vessels,sample_pts[i,:])
        if np.any(dist<R):
            collisions += 10e8*tanh((R - min(dist))*(1/10))
    return collisions


@nb.jit(nopython=True)
def close_exact(data,point):
    line_direction = np.zeros((data.shape[0],3))
    ss = np.zeros(data.shape[0])
    tt = np.zeros(data.shape[0])
    #decision = np.zeros((data.shape[0]))
    hh = np.zeros(data.shape[0])
    cc = np.zeros((data.shape[0],3))
    cd = np.zeros(data.shape[0])
    line_distances = np.zeros(data.shape[0])
    for i in range(data.shape[0]):
        line_direction[i,:] = (data[i,3:6] - data[i,0:3])/np.linalg.norm(data[i,3:6] - data[i,0:3])
        ss[i] = np.dot(data[i,0:3]-point,line_direction[i,:])
        tt[i] = np.dot(point-data[i,3:6],line_direction[i,:])
        d = np.array([ss[i],tt[i],0])
        hh[i] = np.max(d)
        diff = point - data[i,0:3]
        cc[i,:] = np.cross(diff,line_direction[i,:])
        cd[i] = np.linalg.norm(cc[i,:])
        line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i,6]
    return line_distances

def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
    #P1 = np.random.random(3)
    #P2 = np.random.random(3)
    """
    collision_vessels = np.array([[0,0.5,0.25,1,0.5,0.25,0.01],
                                  [0,0.7,0.25,1,0.7,0.25,0.01],
                                  [0,0.3,0.25,1,0.3,0.25,0.01],
                                  [0,0.9,0.25,1,0.9,0.25,0.01],
                                  [0,0.1,0.25,1,0.1,0.25,0.01],
                                  [0.25,0,0.25,0.25,1,0.25,0.01],
                                  [0.75,0,0.25,0.75,1,0.25,0.01],
                                  [0,0.5,0.75,1,0.5,0.75,0.01],
                                  [0,0.7,0.75,1,0.7,0.75,0.01],
                                  [0,0.3,0.75,1,0.3,0.75,0.01],
                                  [0,0.9,0.75,1,0.9,0.75,0.01],
                                  [0,0.1,0.75,1,0.1,0.75,0.01],
                                  [0.25,0,0.75,0.25,1,0.75,0.01],
                                  [0.75,0,0.75,0.75,1,0.75,0.01],
                                  [1,0,0,1,0,1,0.1]])
    """
    #print(collision_vessels.shape)
    #P1 = np.array([0.5,0,0.5])
    #print(close_exact(collision_vessels,P1))
    #P2 = np.array([0.5,0.5,0.5])
    V1 = P2-P1
    # Tangent Constraint 1 for First Spline
    V1 = V1/np.linalg.norm(V1)
    #P3 = np.random.random(3)
    #P4 = np.random.random(3)
    #P3 = np.array([0.5,1,0])
    #P4 = np.array([0.5,0.5,0])
    V2 = P4-P3
    # Tangent Constraint 1 for Second Spline
    V2 = V2/np.linalg.norm(V2)
    MID = (P2 + P4)/2
    V3  = MID - P2
    # Tangent Constraint 2 for First Spline
    V3  = V3/np.linalg.norm(V3)
    V4  = MID - P4
    # Tangent Constraint 2 for Second Spline
    V4  = V4/np.linalg.norm(V4)
    LENGTH = np.linalg.norm(P2-MID)
    CTR0 = P2 + V1*0.01*LENGTH
    CTR1 = P2 + V1*0.1*LENGTH
    CTR2 = P4 + V2*0.1*LENGTH
    CTR3 = (P2+P4)/2 + (V1+V2)/2*0.25*LENGTH
    CTR4 = P4 + V2*0.01*LENGTH
    #CTR3 = P4 + V2*0.1
    # Create Cubic Spline 1
    def get_spline(data,P2=P2,P4=P4,V1=V1,V2=V2,CTR0=CTR0,CTR1=CTR1,CTR2=CTR2,CTR3=CTR3,MID=MID,LENGTH=LENGTH):
        CTR = np.zeros(len(data)).reshape(-1,3)
        seg = np.linspace(0.1,0.9)
        for i in range(CTR.shape[0]):
            CTR[i,:] = P2*(1-seg[i]) + V1*(1-seg[i])*LENGTH + P4*(seg[i]) + V2*(seg[i])*LENGTH
        CTR = CTR + data.reshape(-1,3)
        SPT1 = np.vstack((P2,CTR0,CTR,CTR4,P4))
        try:
            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=2)
        except:
            CTR1 = CTR1 - data[0:3]
            CTR2 = CTR2 - data[3:]
            SPT1 = np.vstack((P2,CTR0,CTR1,CTR2,CTR4,P4))
            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=2)
        spline_f1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=0))
        spline_df1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=1))
        spline_ddf1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=2))
        return spline_f1,spline_df1,spline_ddf1
    def show(ax,spl,color,label,R,P1=P1,P2=P2,P4=P4,CTR4=CTR4,MID=MID):
        #fig = plt.figure()
        #ax = fig.add_subplot(111,projection='3d')
        tt = np.linspace(0,1,num=400)
        spline1 = spl(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(spline1,axis=0),axis=1))
        num = int(spline_length // R) #(R*2*np.pi))
        t = np.linspace(0,1,num=num+2)
        spline1 = spl(t).T
        ax.plot(spline1[:,0],spline1[:,1],spline1[:,2],color=color,label=label)
        ax.plot([P1[0],P2[0]],[P1[1],P2[1]],[P1[2],P2[2]],color='green',label='inlet')
        ax.plot([P3[0],P4[0]],[P3[1],P4[1]],[P3[2],P4[2]],color='green',label='outlet')
        #ax.scatter(SPT1[:,0],SPT1[:,1],SPT1[:,2],color='black',label='data')
        #ax.legend()
        #plt.show()
        return ax
    angle_threshold = 120
    init = np.zeros(9)
    #init[0:3] = CTR1
    #init[3:]  = CTR2
    def cost(data,grad,spline_gen=get_spline,angle_threshold=angle_threshold,R=R,P1=P1,CTR4=CTR4,collision_vessels=collision_vessels):
        spf,spdf,spddf = get_spline(data)
        tt = np.linspace(0,1,num=500)
        pts = spf(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
        num = int(spline_length // R) #(R*2*np.pi))
        sample_tt = np.linspace(0,1,num=4*num+2)
        sample_pts = spf(sample_tt).T
        sample_pts = np.vstack((P1,sample_pts,CTR4))
        vec = np.diff(sample_pts,axis=0)
        collisions = get_collisions(collision_vessels,R,sample_pts)
        vec = vec/np.linalg.norm(vec,axis=0)
        angles = []
        div = 0
        angles = get_vecs(vec)
        a_sum = np.exp(np.sum(angles))
        a_try = np.sum(angles)
        return collisions+a_sum*spline_length #a_sum+spline_length + collisions#+np.tanh(num)
    def cost_polish(data,grad,spline_gen=get_spline,angle_threshold=angle_threshold,R=R,P1=P1,CTR4=CTR4,collision_vessels=collision_vessels):
        spf,spdf,spddf = get_spline(data)
        tt = np.linspace(0,1,num=500)
        pts = spf(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
        num = int(spline_length // R) #(R*2*np.pi))
        sample_tt = np.linspace(0,1,num=4*num+2)
        sample_pts = spf(sample_tt).T
        sample_pts = np.vstack((P1,sample_pts,CTR4))
        vec = np.diff(sample_pts,axis=0)
        collisions = get_collisions(collision_vessels,R,sample_pts)
        vec = vec/np.linalg.norm(vec,axis=0)
        angles = []
        div = 0
        angles = get_vecs(vec)
        a_sum = np.exp(np.sum(angles))
        return collisions+a_sum
    def check(data,spline_gen=get_spline,R=R,P1=P1,CTR4=CTR4):
        spf,spdf,spddf = get_spline(data)
        tt = np.linspace(0,1,num=500)
        pts = spf(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
        num = int(spline_length // R) #(R*2*np.pi))
        #print(num)
        sample_tt = np.linspace(0,1,num=num+2)
        sample_pts = spf(sample_tt).T
        sample_pts = np.vstack((P1,sample_pts,CTR4))
        vec = np.diff(sample_pts,axis=0)
        print(np.linalg.norm(vec,axis=0))
        vec = vec/np.linalg.norm(vec,axis=0)
        #print(vec)
        angles = []
        div = 0
        for i in range(vec.shape[0]-1):
            value = get_angle(vec[i,:],vec[i+1,:])
            angles.append(value)
        return max(angles),min(angles)
    def get_points(data,spline_gen=get_spline):
        spf,spdf,spddf = get_spline(data)
        tt = np.linspace(0,1,num=500)
        pts = spf(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
        num = int(spline_length // R) #(R*2*np.pi))
        #print(num)
        sample_tt = np.linspace(0,1,num=num+2)
        sample_pts = spf(sample_tt).T
        return sample_pts
    def get_curvature(data):
        spf,spdf,spddf = get_spline(data)
        curve = lambda t: curvature(t,df=spdf,ddf=spddf)
        return curve
    def cost2(data,R=R,get_spline=get_spline):
        spf,spdf,spddf = get_spline(data)
        tt = np.linspace(0,1,num=500)
        pts = spf(tt).T
        spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
        kappa = get_curvature(data)
        tt = np.linspace(0,1,num=100)
        k = []
        for t in tt:
            value = 1/kappa(t)
            if value > R:
                k.append(1)
            else:
                k.append(0)
        return np.sum(k)*spline_length
    def cost3(t_vector,grad,spline=None,angle_threshold=angle_threshold,P1=P1,P2=P2,P3=P3,MID=MID,CTR4=CTR4):
        tt = np.linspace(0,1,num=len(t_vector)+2)
        tt[1:-1] = t_vector
        sample_pts = spline(t_vector).T
        sample_pts = np.vstack((P1,sample_pts,P3))
        vec = np.diff(sample_pts,axis=0)
        vec = vec/np.linalg.norm(vec,axis=0)
        angles = []
        for i in range(vec.shape[0]-1):
            value = get_angle(vec[i,:],vec[i+1,:])
            if value < angle_threshold:
                angles.append(1/value)
            else:
                angles.append(0)
        a_sum = np.exp(100*np.sum(angles))
        return a_sum
    return get_spline,show,cost,init,LENGTH,check,get_points,get_curvature,cost3,collision_vessels,cost_polish


def form_connection(R):
    f,s,c,i,L,ch,gp,cv = opt(P1,P2,P3,P4,R)
    spf,spdf = f(i)
    res1 = optimize.shgo(c,iters=3,bounds=[(-L,L),(-L,L),(-L,L),(-L,L),(-L,L),(-L,L)],minimizer_kwargs={"method":"Nelder-Mead","N":12000})
    spf,spdf = f(res1.x)
    pts = gp(res1.x)
    return pts

def test_nlopt(R):
    f,s,c,i,L,ch,gp,gc,c3,cv,cp = opt(R)
    fig = plt.figure()
    ax1 = fig.add_subplot(111,projection='3d')
    #ax2 = fig.add_subplot(122,projection='3d')
    spf,spdf,spddf = f(i)
    ax1 = s(ax1,spf,'red','initial',R)
    for i in range(cv.shape[0]):
        ax1.plot([cv[i,0],cv[i,3]],[cv[i,1],cv[i,4]],[cv[i,2],cv[i,5]],color='black',label='other vessel {}'.format(i))
    success = False
    count = 0
    max_time = 25
    x0 = np.zeros((count+2)*3)
    while not success:
        print('Linking Optimization Path Search {}'.format(count))
        bounds = [(-L,L)]*((count+2)*3)
        lb = np.ones((count+2)*3)*(-L) + x0
        ub = np.ones((count+2)*3)*(L) + x0
        o = nlopt.opt(2,(count+2)*3)
        o.set_min_objective(c)
        o.set_lower_bounds(lb)
        o.set_upper_bounds(ub)
        o.set_maxtime(max_time)
        xopt = o.optimize(x0)
        if o.last_optimum_value() < 10:
            L += L*0.25
            success = True
        else:
            print("Above link threshold {}".format(o.last_optimum_value()))
            L += L*0.25
            count += 1
            max_time += 0
            x0 = np.zeros((count+2)*3)
            spfc,spdfc,spddfc = f(xopt)
            tj = np.linspace(0.1,0.9,num=count+2)
            for j in range(len(tj)):
                x0[j*3:j*3+3] = spfc(tj[j])
    spfc,spdfc,spddfc = f(xopt)
    print("Collision Optimum: {}".format(o.last_optimum_value()))
    old_best = cp(xopt,np.array([]))
    tt = np.linspace(0,1,num=500)
    pts = spfc(tt).T
    ax1.plot(pts[:,0],pts[:,1],pts[:,2],color='orange',label='no collision spline')
    for i in range(int(xopt.size//3)):
        def local_cost(point,grad,idx=i,data=xopt,lc=cp):
            data[i*3:i*3+3] = point
            return c(data,grad)
        o = nlopt.opt(2,3)
        o.set_min_objective(local_cost)
        o.set_lower_bounds(lb[i*3:i*3+3])
        o.set_upper_bounds(ub[i*3:i*3+3])
        o.set_maxtime(15)
        local_xopt = o.optimize(xopt[i*3:i*3+3])
        if o.last_optimum_value() < old_best:
            xopt[i*3:i*3+3] = local_xopt
            print('New Optimum Found')
            old_best = o.last_optimum_value()
    print('Finished Local Adjustments')
    spf,spdf,spddf = f(xopt)
    print("Curvature Optimum: {}".format(o.last_optimum_value()))
    cost3 = lambda t,grad: c3(t,grad,spline=spf)
    tt = np.linspace(0,1,num=500)
    pts = spf(tt).T
    ax1.plot(pts[:,0],pts[:,1],pts[:,2],color='yellow',label='curvature optimized spline')
    pts_xopt = xopt.reshape(-1,3)
    ax1.scatter(pts_xopt[:,0],pts_xopt[:,1],pts_xopt[:,2],color='red')
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    num = int(spline_length // R)
    t_vector = np.linspace(0,1,num=num+2)
    bounds = []
    lb = []
    ub = []
    for i in range(len(t_vector)-1):
        lb.append(t_vector[i])
        ub.append(t_vector[i+1])
        bounds.append(tuple([t_vector[i],t_vector[i+1]]))
    lb = np.array(lb)
    ub = np.array(ub)
    t0 = (lb + ub)/2
    print("Performing Vessel Annealing Optimization")
    seg_opt = nlopt.opt(2,len(t0))
    seg_opt.set_min_objective(cost3)
    seg_opt.set_lower_bounds(lb)
    seg_opt.set_upper_bounds(ub)
    seg_opt.set_stopval(100)
    seg_opt.set_maxtime(100)
    t_vector = seg_opt.optimize(t0)
    tt = np.linspace(0,1,num=len(t_vector)+2)
    tt[1:-1] = t_vector
    spline1 = spf(tt).T
    ax1.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='final')
    plt.show()
    return xopt,spf


def test(R):
    f,s,c,i,L,ch,gp,gc,c3,cv,cp = opt(R)
    fig = plt.figure()
    ax1 = fig.add_subplot(111,projection='3d')
    #ax2 = fig.add_subplot(122,projection='3d')
    spf,spdf,spddf = f(i)
    ax1 = s(ax1,spf,'red','initial',R)
    for i in range(cv.shape[0]):
        ax1.plot([cv[i,0],cv[i,3]],[cv[i,1],cv[i,4]],[cv[i,2],cv[i,5]],color='black',label='other vessel {}'.format(i))
    #ax2 = s(ax2,spf,'red','initial',R)
    #res = optimize.minimize(c,i,method='Nelder-Mead',options={'disp':True})
    #res = optimize.differential_evolution(c,maxiter=8000,bounds=[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)])
    #print("Performing Brute Seed Selection")
    #x0 = optimize.brute(c,Ns=5,ranges=[(-L,L),(-L,L),(-L,L),(-L,L),(-L,L),(-L,L)],disp=True)
    #print("Basinhopping Optimization")
    #res1 = optimize.basinhopping(c,x0)
    #print("Spline Optimization Complete")
    success = False
    count = 0
    while not success:
        print('Linking Optimization Path Search {}'.format(count))
        bounds = [(-L,L)]*((count+2)*3)
        res1 = optimize.shgo(c,iters=1,bounds=bounds,options={'disp':True})#,minimizer_kwargs={"method":"Nelder-Mead","N":12000})
        #res1 = optimize.dual_annealing(c,bounds=bounds)
        #res1 = optimize.basinhopping(c,np.zeros((count+2)*3))
        #res1 = optimize.differential_evolution(c,bounds=bounds)
        #res1 = optimize.minimize(c,np.zeros((count+2)*3),method="L-BFGS-B",bounds=bounds)
        if res1.fun < 10:
            L += L*0.25
            bounds = [(-L,L)]*((count+2)*3)
            success = True
        else:
            print("Above link threshold {}".format(res1.fun))
            L += L*0.25
            count += 1
    res1 = optimize.minimize(cp,res1.x,method='L-BFGS-B',bounds=bounds,options={'disp':1})
    #res2 = optimize.basinhopping(c,np.zeros(6),niter=200,minimizer_kwargs={"method":"Nelder-Mead"})#,bounds=[(-L,L),(-L,L),(-L,L),(-L,L),(-L,L),(-L,L)])#,minimizer_kwargs={"method":"BFGS"})
    #res = optimize.basinhopping(c,i,minimizer_kwargs={"method": "L-BFGS-B","bounds":[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)]},T=0.25,niter=200,disp=True) #,bounds=[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)])#,i,minimizer_kwargs={"method": "Nelder-Mead"},disp=True)#bounds=[(0,1),(0,1),(0,1),(0,1),(0,1),(0,1)])
    spf,spdf,spddf = f(res1.x)
    print("Curvature Optimum: {}".format(res1.fun))
    cost3 = lambda t: c3(t,spline=spf)
    tt = np.linspace(0,1,num=500)
    pts = spf(tt).T
    ax1.plot(pts[:,0],pts[:,1],pts[:,2],color='yellow',label='intermediate')
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    num = int(spline_length // R)
    print(num)
    t_vector = np.linspace(0,1,num=num+2)
    bounds = []
    for i in range(len(t_vector)-1):
        bounds.append(tuple([t_vector[i],t_vector[i+1]]))
    t0 = t_vector[1:-1]
    print("Performing Vessel Annealing Optimization")
    res3 = optimize.dual_annealing(cost3,bounds=bounds)#,minimizer_kwargs={"method":"L-BFGS-B"})
    t_vector = res3.x
    tt = np.linspace(0,1,num=len(t_vector)+2)
    tt[1:-1] = t_vector
    spline1 = spf(tt).T
    ax1.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='final')
    #ax1 = s(ax1,spf,'blue','final',R)
    #ax1.set_title('results 1')
    #spf,spdf = f(res2.x)
    #ax2 = s(ax2,spf,'blue','final',R)
    #ax2.set_title('results 2')
    #plt.legend()
    plt.show()
    #print(res.fun)
    #ANGLES = ch(res1.x)
    #print('Min angle:\n{}'.format(ANGLES[1]))
    #print('Max angle:\n{}'.format(ANGLES[0]))
    #ANGLES = ch(res2.x)
    #print('Min angle:\n{}'.format(ANGLES[1]))
    #print('Max angle:\n{}'.format(ANGLES[0]))
    pts = gp(res1.x)
    return pts


def retest(data,f,s,c,i,L,ch):
    fig = plt.figure()
    ax1 = fig.add_subplot(121,projection='3d')
    ax2 = fig.add_subplot(122,projection='3d')
    spf,spdf = f(i)
    ax1 = s(ax1,spf,'red','initial',R)
    ax2 = s(ax2,spf,'red','initial',R)
    #res = optimize.minimize(c,i,method='Nelder-Mead',options={'disp':True})
    #res = optimize.differential_evolution(c,maxiter=8000,bounds=[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)])
    res1 = optimize.shgo(c,iters=3,bounds=[(-L,L),(-L,L),(-L,L),(-L,L),(-L,L),(-L,L)],minimizer_kwargs={"method":"Nelder-Mead","N":12000})
    res2 = optimize.shgo(c,iters=3,bounds=[(-L,L),(-L,L),(-L,L),(-L,L),(-L,L),(-L,L)],minimizer_kwargs={"method":"BFGS"})
    #res = optimize.basinhopping(c,i,minimizer_kwargs={"method": "L-BFGS-B","bounds":[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)]},T=0.25,niter=200,disp=True) #,bounds=[(-1,1),(-1,1),(-1,1),(-1,1),(-1,1),(-1,1)])#,i,minimizer_kwargs={"method": "Nelder-Mead"},disp=True)#bounds=[(0,1),(0,1),(0,1),(0,1),(0,1),(0,1)])
    spf,spdf = f(res1.x)
    ax1 = s(ax1,spf,'blue','final',R)
    ax1.set_title('results 1')
    spf,spdf = f(res2.x)
    ax2 = s(ax2,spf,'blue','final',R)
    ax2.set_title('results 2')
    plt.show()
    #print(res.fun)
    ANGLES = ch(res1.x)
    print('Min angle:\n{}'.format(ANGLES[1]))
    print('Max angle:\n{}'.format(ANGLES[0]))
    ANGLES = ch(res2.x)
    print('Min angle:\n{}'.format(ANGLES[1]))
    print('Max angle:\n{}'.format(ANGLES[0]))

def find(spf):
    inital_length

def show(spl):
    fig = plt.figure()
    ax = fig.add_subplot(111,projection='3d')
    tt = np.linspace(0,1,num=400)
    spline1 = spl(tt).T
    ax.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='unclamped')
    #ax.plot([P1[0],P2[0]],[P1[1],P2[1]],[P1[2],P2[2]],color='green',label='inlet')
    #ax.plot([MID[0],CTR4[0]],[MID[1],CTR4[1]],[MID[2],CTR4[2]],color='green',label='outlet')
    #ax.scatter(SPT1[:,0],SPT1[:,1],SPT1[:,2],color='black',label='data')
    #ax.legend()
    plt.show()
